@@ -3,18 +3,32 @@
 import os
 import re
 import ctypes as ct
+import shlex
 import socket
 import struct
+import sys
 import weakref
 
-from util import find_library
+from util import find_library, load_kernel
 from xtables import (XT_INV_PROTO, NFPROTO_IPV4, XTablesError, xtables,
                      xt_align, xt_counters, xt_entry_target, xt_entry_match)
 
 __all__ = ["Table", "Chain", "Rule", "Match", "Target", "Policy", "IPTCError"]
 
+load_kernel("ip_tables")
+
 _IFNAMSIZ = 16
 
+_libc = ct.CDLL("libc.so.6")
+_get_errno_loc = _libc.__errno_location
+_get_errno_loc.restype = ct.POINTER(ct.c_int)
+_malloc = _libc.malloc
+_malloc.restype = ct.POINTER(ct.c_ubyte)
+_malloc.argtypes = [ct.c_size_t]
+_free = _libc.free
+_free.restype = None
+_free.argtypes = [ct.POINTER(ct.c_ubyte)]
+
 
 def is_table_available(name):
     try:
@@ -234,7 +248,7 @@ class iptc(object):
 class IPTCModule(object):
     """Superclass for Match and Target."""
     pattern = re.compile(
-        '\s*(!)?\s*--([-\w]+)\s+(!)?\s*"?([^"]*?)"?(?=\s*(?:!?\s*--|$))')
+        '\s*(!)?\s*--([-\w]+)\s+(!)?\s*("?[^"]*?"?)(?=\s*(?:!?\s*--|$))')
 
     def __init__(self):
         self._name = None
@@ -258,9 +272,14 @@ class IPTCModule(object):
         else:
             inv = ct.c_int(0)
 
-        argv = (ct.c_char_p * 2)()
+        args = shlex.split(value)
+        if not args:
+            args = [value]
+        N = len(args)
+        argv = (ct.c_char_p * (N + 1))()
         argv[0] = parameter
-        argv[1] = value
+        for i in xrange(N):
+            argv[i + 1] = args[i]
 
         entry = self._rule.entry and ct.pointer(self._rule.entry) or None
 
@@ -276,27 +295,40 @@ class IPTCModule(object):
     def _final_check(self):
         raise NotImplementedError()
 
+    def _get_saved_buf(self, ip):
+        if not self._module or not self._module.save:
+            return None
+        # redirect C stdout to a pipe and read back the output of m->save
+        fd = sys.stdout.fileno()
+        with os.fdopen(os.dup(fd), 'w') as old_stdout:
+            try:
+                pipes = os.pipe()
+                sys.stdout.close()
+                os.dup2(pipes[1], fd)
+                sys.stdout = os.fdopen(fd, 'w')
+                self._xt.save(self._module, ip, self._ptr)
+                buf = os.read(pipes[0], 1024)
+                os.close(pipes[0])
+                os.close(pipes[1])
+                return buf
+            finally:
+                sys.stdout.close()
+                os.dup2(old_stdout.fileno(), fd)
+                sys.stdout = os.fdopen(fd, 'w')
+
     def save(self, name):
         return self._save(name, self.rule.get_ip())
 
     def _save(self, name, ip):
-        if self._module and self._module.save:
-            # redirect C stdout to a pipe and read back the output of m->save
-            pipes = os.pipe()
-            saved_out = os.dup(1)
-            os.dup2(pipes[1], 1)
-            self._xt.save(self._module, ip, self._ptr)
-            buf = os.read(pipes[0], 1024)
-            os.dup2(saved_out, 1)
-            os.close(pipes[0])
-            os.close(pipes[1])
-            os.close(saved_out)
-            if name:
-                return self._get_value(buf, name)
-            else:
-                return self._get_all_values(buf)
-        else:
+        buf = self._get_saved_buf(ip)
+        if buf is None:
             return None
+        if not self._module or not self._module.save:
+            return None
+        if name:
+            return self._get_value(buf, name)
+        else:
+            return self._get_all_values(buf)
 
     def _get_all_values(self, buf):
         table = {}  # variable -> (value, inverted)
@@ -321,22 +353,11 @@ class IPTCModule(object):
     def get_all_parameters(self):
         params = {}
         ip = self.rule.get_ip()
-        if self._module and self._module.save:
-            # redirect C stdout to a pipe and read back the output of m->save
-            pipes = os.pipe()
-            saved_out = os.dup(1)
-            os.dup2(pipes[1], 1)
-            self._xt.save(self._module, ip, self._ptr)
-            buf = os.read(pipes[0], 1024)
-            os.dup2(saved_out, 1)
-            os.close(pipes[0])
-            os.close(pipes[1])
-            os.close(saved_out)
-
+        buf = self._get_saved_buf(ip)
+        if buf is not None:
             res = re.findall(IPTCModule.pattern, buf)
             for x in res:
                 params[x[1]] = "%s%s" % ((x[0] or x[2]) and "!" or "", x[3])
-
         return params
 
     def _update_parameters(self):
@@ -374,6 +395,20 @@ class IPTCModule(object):
     """The rule this target or match belong to."""
 
 
+class _Buffer(object):
+    def __init__(self, size=0):
+        if size > 0:
+            self.buffer = _malloc(size)
+            if self.buffer is None:
+                raise Exception("Can't allocate buffer")
+        else:
+            self.buffer = None
+
+    def __del__(self):
+        if self.buffer is not None:
+            _free(self.buffer)
+
+
 class Match(IPTCModule):
     """Matches are extensions which can match for special header fields or
     other attributes of a packet.
@@ -408,6 +443,8 @@ class Match(IPTCModule):
             name = match.u.user.name
         self._name = name
         self._rule = rule
+        self._alias = None
+        self._real_name = None
 
         self._xt = xtables(rule.nfproto)
 
@@ -420,6 +457,10 @@ class Match(IPTCModule):
             self._revision = revision
         else:
             self._revision = self._module.revision
+        if self._module.next is not None:
+            self._store_buffer(module)
+
+        self._check_alias(module[0], match)
 
         self._match_buf = (ct.c_ubyte * self.size)()
         if match:
@@ -439,14 +480,37 @@ class Match(IPTCModule):
             return True
         return False
 
-    def __ne__(self, rule):
-        return not self.__eq__(rule)
+    def __ne__(self, match):
+        return not self.__eq__(match)
+
+    def _check_alias(self, module, match):
+        # This is ugly, but there are extensions using an alias name. Check if
+        # that's the case, and load that extension as well if necessary. It
+        # will be used to parse parameters, since the 'real' extension
+        # probably won't understand them.
+        if getattr(module, "real_name", None) is not None and module.real_name:
+            self._real_name = module.real_name
+        if getattr(module, "alias", None) is not None and module.alias:
+            self._alias_name = module.alias(match)
+            alias = self._xt.find_match(self._alias_name)
+            if not alias:
+                raise XTablesError("can't find alias match %s" %
+                                   (self._alias_name))
+            self._alias = alias[0]
+
+    def _store_buffer(self, module):
+        self._buffer = _Buffer()
+        self._buffer.buffer = ct.cast(module, ct.POINTER(ct.c_ubyte))
 
     def _final_check(self):
         self._xt.final_check_match(self._module)
 
     def _parse(self, argv, inv, entry):
-        self._xt.parse_match(argv, inv, self._module, entry,
+        if self._alias is not None:
+            module = self._alias
+        else:
+            module = self._module
+        self._xt.parse_match(argv, inv, module, entry,
                              ct.cast(self._ptrptr, ct.POINTER(ct.c_void_p)))
 
     def _get_size(self):
@@ -466,6 +530,16 @@ class Match(IPTCModule):
         self._ptrptr = ct.cast(ct.pointer(self._ptr),
                                ct.POINTER(ct.POINTER(xt_entry_match)))
         self._module.m = self._ptr
+        if self._alias is not None:
+            self._alias.m = self._ptr
+        self._update_name()
+
+    def _update_name(self):
+        m = self._ptr[0]
+        if self._real_name is not None:
+            m.u.user.name = self._real_name
+        else:
+            m.u.user.name = self.name
 
     def reset(self):
         """Reset the match.
@@ -474,7 +548,6 @@ class Match(IPTCModule):
         ct.memset(ct.byref(self._match_buf), 0, self.size)
         self._update_pointers()
         m = self._ptr[0]
-        m.u.user.name = self.name
         m.u.match_size = self.size
         m.u.user.revision = self._revision
         if self._module.init:
@@ -528,17 +601,11 @@ class Target(IPTCModule):
 
         self._xt = xtables(rule.nfproto)
 
-        is_standard_target = False
-        module = None
-        for t in rule.tables:
-            if t.is_chain(name):
-                is_standard_target = True
-                module = self._xt.find_target('standard')
-
+        module = (self._is_standard_target() and
+                  self._xt.find_target('standard') or
+                  self._xt.find_target(name))
         if not module:
-            module = self._xt.find_target(name)
-            if not module:
-                raise XTablesError("can't find target %s" % (name))
+            raise XTablesError("can't find target %s" % (name))
         self._module = module[0]
         self._module.tflags = 0
         if revision is not None:
@@ -546,15 +613,11 @@ class Target(IPTCModule):
         else:
             self._revision = self._module.revision
 
-        self._target_buf = (ct.c_ubyte * self.size)()
-        if target:
-            ct.memmove(ct.byref(self._target_buf), ct.byref(target), self.size)
-            self._update_pointers()
-            self._update_parameters()
-        else:
-            self.reset()
+        self._check_alias(module[0], target)
 
-        if is_standard_target:
+        self._create_buffer(target)
+
+        if self._is_standard_target():
             self.standard_target = name
 
     def __eq__(self, targ):
@@ -570,20 +633,58 @@ class Target(IPTCModule):
             self.target.u.user.name == "RETURN" or
             self.target.u.user.name == "ERROR"):
             return True
-        if (self.target_buf[basesz:self.usersize] ==
-            targ.target_buf[basesz:targ.usersize]):
+        if (self._target_buf[basesz:self.usersize] ==
+            targ._target_buf[basesz:targ.usersize]):
             return True
         return False
 
-    def __ne__(self, rule):
-        return not self.__eq__(rule)
+    def __ne__(self, target):
+        return not self.__eq__(target)
+
+    def _check_alias(self, module, target):
+        # This is ugly, but there are extensions using an alias name. Check if
+        # that's the case, and load that extension as well if necessary. It
+        # will be used to parse parameters, since the 'real' extension
+        # probably won't understand them.
+        if getattr(module, "real_name", None) is not None and module.real_name:
+            self._real_name = module.real_name
+        if getattr(module, "alias", None) is not None and module.alias:
+            self._alias_name = module.alias(target)
+            alias = self._xt.find_target(self._alias_name)
+            if not alias:
+                raise XTablesError("can't find alias target %s" %
+                                   (self._alias_name))
+            self._alias = alias[0]
+
+    def _create_buffer(self, target):
+        self._buffer = _Buffer(self.size)
+        self._target_buf = self._buffer.buffer
+        if target:
+            ct.memmove(self._target_buf, ct.byref(target), self.size)
+            self._update_pointers()
+            self._update_parameters()
+        else:
+            self.reset()
+
+    def _is_standard_target(self):
+        for t in self._rule.tables:
+            if t.is_chain(self._name):
+                return True
+        return False
 
     def _final_check(self):
         self._xt.final_check_target(self._module)
 
     def _parse(self, argv, inv, entry):
-        self._xt.parse_target(argv, inv, self._module, entry,
+        if self._alias is not None:
+            module = self._alias
+        else:
+            module = self._module
+        self._xt.parse_target(argv, inv, module, entry,
                               ct.cast(self._ptrptr, ct.POINTER(ct.c_void_p)))
+        self._target_buf = ct.cast(self._module.t, ct.POINTER(ct.c_ubyte))
+        self._buffer.buffer = self._target_buf
+        self._update_pointers()
 
     def _get_size(self):
         return xt_align(self._module.size + ct.sizeof(xt_entry_target))
@@ -610,19 +711,27 @@ class Target(IPTCModule):
     into."""
 
     def _update_pointers(self):
-        self._ptr = ct.cast(ct.byref(self._target_buf),
-                            ct.POINTER(xt_entry_target))
+        self._ptr = ct.cast(self._target_buf, ct.POINTER(xt_entry_target))
         self._ptrptr = ct.cast(ct.pointer(self._ptr),
                                ct.POINTER(ct.POINTER(xt_entry_target)))
         self._module.t = self._ptr
+        if self._alias is not None:
+            self._alias.t = self._ptr
+        self._update_name()
+
+    def _update_name(self):
+        m = self._ptr[0]
+        if self._real_name is not None:
+            m.u.user.name = self._real_name
+        else:
+            m.u.user.name = self.name
 
     def reset(self):
         """Reset the target.  Parameters are set to their default values, any
         flags are cleared."""
-        ct.memset(ct.byref(self._target_buf), 0, self.size)
+        ct.memset(self._target_buf, 0, self.size)
         self._update_pointers()
         t = self._ptr[0]
-        t.u.user.name = self.name
         t.u.target_size = self.size
         t.u.user.revision = self._revision
         if self._module.init:
@@ -630,16 +739,10 @@ class Target(IPTCModule):
         self._module.tflags = 0
 
     def _get_target(self):
-        return ct.cast(ct.byref(self.target_buf),
-                       ct.POINTER(xt_entry_target))[0]
+        return self._ptr[0]
     target = property(_get_target)
     """This is the C structure used by the extension."""
 
-    def _get_target_buf(self):
-        return self._target_buf
-    target_buf = property(_get_target_buf)
-    """This is the buffer holding the C structure used by the extension."""
-
 
 class Policy(object):
     """
@@ -1282,18 +1385,21 @@ class Table(object):
     ALL = ["filter", "mangle", "raw", "nat"]
     """This is the constant for all tables."""
 
-    _cache = weakref.WeakValueDictionary()
+    _cache = dict()
 
-    def __new__(cls, name, autocommit=True):
+    def __new__(cls, name, autocommit=None):
         obj = Table._cache.get(name, None)
         if not obj:
             obj = object.__new__(cls)
+            if autocommit is None:
+                autocommit = True
+            obj._init(name, autocommit)
             Table._cache[name] = obj
-        else:
+        elif autocommit is not None:
             obj.autocommit = autocommit
         return obj
 
-    def __init__(self, name, autocommit=True):
+    def _init(self, name, autocommit):
         """
         *name* is the name of the table, if it already exists it is returned.
         *autocommit* specifies that any iptables operation that changes a
@@ -1323,7 +1429,8 @@ class Table(object):
         if self._handle is None:
             raise IPTCError("table is not initialized")
         try:
-            self.commit()
+            if self.autocommit:
+                self.commit()
         except IPTCError, e:
             if not ignore_exc:
                 raise e
@@ -1362,8 +1469,7 @@ class Table(object):
 
     def strerror(self):
         """Returns any pending iptables error from the previous operation."""
-        errno = ct.get_errno()
-        ct.set_errno(0)
+        errno = _get_errno_loc()[0]
         if errno == 0:
             return "libiptc version error"
         return self._iptc.iptc_strerror(errno)
