@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 
 import ctypes as ct
+import os
 import sys
 import weakref
 import version
@@ -385,8 +386,7 @@ class _xtables_match_v10(ct.Structure):
                 ("save", ct.CFUNCTYPE(None, ct.c_void_p,
                                       ct.POINTER(xt_entry_match))),
                 # Print match name or alias
-                ("alias", ct.CFUNCTYPE(ct.c_char_p, ct.c_void_p,
-                                       ct.POINTER(xt_entry_match))),
+                ("alias", ct.CFUNCTYPE(ct.c_char_p, ct.POINTER(xt_entry_match))),
                 # pointer to list of extra command-line options
                 ("extra_opts", ct.POINTER(option)),
 
@@ -635,8 +635,7 @@ class _xtables_target_v10(ct.Structure):
                 ("save", ct.CFUNCTYPE(None, ct.c_void_p,
                                       ct.POINTER(xt_entry_target))),
                 # Print target name or alias
-                ("alias", ct.CFUNCTYPE(ct.c_char_p, ct.c_void_p,
-                                       ct.POINTER(xt_entry_target))),
+                ("alias", ct.CFUNCTYPE(ct.c_char_p, ct.POINTER(xt_entry_target))),
                 # pointer to list of extra command-line options
                 ("extra_opts", ct.POINTER(option)),
 
@@ -670,16 +669,6 @@ class xtables_target(ct.Union):
                 ("v10", _xtables_target_v10)]
 
 
-class xtables_afinfo(ct.Structure):
-    _fields_ = [("kmod", ct.c_char_p),
-                ("proc_exists", ct.c_char_p),
-                ("libprefix", ct.c_char_p),
-                ("family", ct.c_uint8),
-                ("ipproto", ct.c_uint8),
-                ("so_rev_match", ct.c_int),
-                ("so_rev_target", ct.c_int)]
-
-
 class XTablesError(Exception):
     """Raised when an xtables call fails for some reason."""
 
@@ -689,6 +678,18 @@ _optind = ct.c_long.in_dll(_libc, "optind")
 _optarg = ct.c_char_p.in_dll(_libc, "optarg")
 
 _lib_xtables, _xtables_version = find_library("xtables")
+_xtables_libdir = os.getenv("XTABLES_LIBDIR")
+if _xtables_libdir is None:
+    import os.path
+    for xtdir in ["/lib/xtables", "/lib64/xtables", "/usr/lib/xtables",
+                  "/usr/lib/iptables", "/usr/lib64/xtables",
+                  "/usr/lib64/iptables", "/usr/local/lib/xtables"]:
+        if os.path.isdir(xtdir):
+            _xtables_libdir = xtdir
+            break
+if _xtables_libdir is None:
+    raise XTablesError("can't find directory with extensions; "
+                       "please set XTABLES_LIBDIR")
 
 _lib_xtwrapper, _ = find_library("xtwrapper")
 
@@ -754,14 +755,23 @@ class xtables(object):
     _xtables_find_target.restype = ct.POINTER(xtables_target)
     _xtables_find_target.argtypes = [ct.c_char_p, ct.c_int]
 
-    _xtables_afinfo = ct.c_void_p.in_dll(_lib_xtables, "afinfo")
+    _xtables_set_nfproto = _lib_xtables.xtables_set_nfproto
+    _xtables_set_nfproto.restype = None
+    _xtables_set_nfproto.argtypes = [ct.c_uint8]
+
     _xtables_xt_params = ct.c_void_p.in_dll(_lib_xtables, "xt_params")
     _xtables_matches = (ct.c_void_p.in_dll(_lib_xtables, "xtables_matches"))
-    _xtables_pending_matches = (ct.c_void_p.in_dll(_lib_xtables,
-                                                   "xtables_pending_matches"))
+    try:
+        _xtables_pending_matches = (ct.c_void_p.in_dll(_lib_xtables,
+                                                       "xtables_pending_matches"))
+    except ValueError:
+        _xtables_pending_matches = ct.POINTER(None)
     _xtables_targets = (ct.c_void_p.in_dll(_lib_xtables, "xtables_targets"))
-    _xtables_pending_targets = (ct.c_void_p.in_dll(_lib_xtables,
-                                                   "xtables_pending_targets"))
+    try:
+        _xtables_pending_targets = (ct.c_void_p.in_dll(_lib_xtables,
+                                                       "xtables_pending_targets"))
+    except ValueError:
+        _xtables_pending_targets = ct.POINTER(None)
 
     _cache = weakref.WeakValueDictionary()
 
@@ -796,7 +806,6 @@ class xtables(object):
         self._loaded_exts = []
 
         # make sure we're initializing with clean state
-        self._afinfo = ct.c_void_p(None).value
         self._xt_params = ct.c_void_p(None).value
         self._matches = ct.c_void_p(None).value
         self._pending_matches = ct.c_void_p(None).value
@@ -815,8 +824,6 @@ class xtables(object):
         # Save our per-protocol libxtables global variables, and set them to
         # NULL so that we don't interfere with other protocols.
         null = ct.c_void_p(None)
-        self._afinfo = xtables._xtables_afinfo.value
-        xtables._xtables_afinfo.value = null.value
         self._xt_params = xtables._xtables_xt_params.value
         xtables._xtables_xt_params.value = null.value
         self._matches = xtables._xtables_matches.value
@@ -831,7 +838,7 @@ class xtables(object):
     def _restore_globals(self):
         # Restore per-protocol libxtables global variables saved in
         # _save_globals().
-        xtables._xtables_afinfo.value = self._afinfo
+        xtables._xtables_set_nfproto(self.proto)
         xtables._xtables_xt_params.value = self._xt_params
         xtables._xtables_matches.value = self._matches
         xtables._xtables_pending_matches.value = self._pending_matches
@@ -856,8 +863,7 @@ class xtables(object):
         try:
             initfn = getattr(lib, "libxt_%s_init" % (name))
         except AttributeError:
-            afinfo = ct.cast(self._afinfo, ct.POINTER(xtables_afinfo))
-            prefix = afinfo[0].libprefix
+            prefix = self._get_prefix()
             initfn = getattr(lib, "%s%s_init" % (prefix, name), None)
         return initfn
 
@@ -873,13 +879,20 @@ class xtables(object):
             pass
         return False
 
+    def _get_prefix(self):
+        if self.proto == NFPROTO_IPV4:
+            return "libipt_"
+        elif self.proto == NFPROTO_IPV6:
+            return "libip6t_"
+        else:
+            raise XTablesError("Unknown protocol %d" % (self.proto))
+
     def _try_register(self, name):
         if self._try_extinit(name, _lib_xtables):
             return
-        afinfo = ct.cast(self._afinfo, ct.POINTER(xtables_afinfo))
-        prefix = afinfo[0].libprefix
-        libs = ["/lib/xtables/libxt_" + name + ".so",
-                "/lib/xtables/" + prefix + name + ".so"]
+        prefix = self._get_prefix()
+        libs = [os.path.join(_xtables_libdir, "libxt_" + name + ".so"),
+                os.path.join(_xtables_libdir, prefix + name + ".so")]
         for lib in libs:
             if self._try_extinit(name, lib):
                 return
@@ -1037,6 +1050,29 @@ class xtables(object):
                     continue
             # XXX: check for conflicting options
 
+    def _fcheck_target_old(self, target):
+        # old API
+        if not target.final_check:
+            return
+        rv = _wrap_uintfn(target.final_check, target.tflags)
+        if rv:
+            raise XTablesError("%s.final_check() has failed" %
+                               (target.name))
+
+    def _fcheck_target_new(self, target):
+        # new API
+        cb = xt_fcheck_call()
+        cb.ext_name = target.name
+        cb.data = ct.cast(target.t[0].data, ct.c_void_p)
+        cb.xflags = target.tflags
+        cb.udata = target.udata
+        rv = _wrap_x6fn(target.x6_fcheck, ct.pointer(cb))
+        if rv:
+            raise XTablesError("%s.x6_fcheck has failed" % (target.name))
+        if target.x6_options:
+            self._options_fcheck(target.name, target.tflags,
+                                 target.x6_options)
+
     # Dispatch arguments to the appropriate final_check function, based upon
     # the extension's choice of API.
     @preserve_globals
@@ -1050,30 +1086,32 @@ class xtables(object):
             pass
 
         if x6_fcheck:
-            # new API
-            cb = xt_fcheck_call()
-            cb.ext_name = target.name
-            cb.data = ct.cast(target.t[0].data, ct.c_void_p)
-            cb.xflags = target.tflags
-            cb.udata = target.udata
-            rv = _wrap_x6fn(target.x6_fcheck, ct.pointer(cb))
-            if rv:
-                raise XTablesError("%s.x6_fcheck has failed" % (target.name))
+            self._fcheck_target_new(target)
         else:
-            if target.final_check:
-                rv = _wrap_uintfn(target.final_check, target.tflags)
-                if rv:
-                    raise XTablesError("%s.final_check() has failed" %
-                                       (target.name))
+            self._fcheck_target_old(target)
 
-        try:
-            # new API
-            if target.x6_options:
-                self._options_fcheck(target.name, target.tflags,
-                                     target.x6_options)
-        except AttributeError:
-            # old API
-            pass
+    def _fcheck_match_old(self, match):
+        # old API
+        if not match.final_check:
+            return
+        rv = _wrap_uintfn(match.final_check, match.mflags)
+        if rv:
+            raise XTablesError("%s.final_check() has failed" %
+                               (match.name))
+
+    def _fcheck_match_new(self, match):
+        # new API
+        cb = xt_fcheck_call()
+        cb.ext_name = match.name
+        cb.data = ct.cast(match.m[0].data, ct.c_void_p)
+        cb.xflags = match.mflags
+        cb.udata = match.udata
+        rv = _wrap_x6fn(match.x6_fcheck, ct.pointer(cb))
+        if rv:
+            raise XTablesError("%s.x6_fcheck has failed" % (match.name))
+        if match.x6_options:
+            self._options_fcheck(match.name, match.mflags,
+                                 match.x6_options)
 
     # Dispatch arguments to the appropriate final_check function, based upon
     # the extension's choice of API.
@@ -1083,32 +1121,11 @@ class xtables(object):
         try:
             # new API?
             x6_fcheck = match.x6_fcheck
-        except AttributeError:  # old API
-            pass
-
-        if x6_fcheck:
-            # new API
-            cb = xt_fcheck_call()
-            cb.ext_name = match.name
-            cb.data = ct.cast(match.m[0].data, ct.c_void_p)
-            cb.xflags = match.mflags
-            cb.udata = match.udata
-            rv = _wrap_x6fn(match.x6_fcheck, ct.pointer(cb))
-            if rv:
-                raise XTablesError("%s.x6_fcheck has failed" % (match.name))
-        else:
-            # old API
-            if match.final_check:
-                rv = _wrap_uintfn(match.final_check, match.mflags)
-                if rv:
-                    raise XTablesError("%s.final_check() has failed" %
-                                       (match.name))
-
-        try:
-            # new API
-            if match.x6_options:
-                self._options_fcheck(match.name, match.mflags,
-                                     match.x6_options)
         except AttributeError:
             # old API
             pass
+
+        if x6_fcheck:
+            self._fcheck_match_new(match)
+        else:
+            self._fcheck_match_old(match)
